global function OnWeaponPrimaryAttack_cloak

const float CLOAK_FIELD_COOLDOWN_TIME = 40

const float CLOAK_DRONE_LIFETIME = 20
const float CLOAK_DRONE_COOLDOWN_TIME = 40

var function OnWeaponPrimaryAttack_cloak( entity weapon, WeaponPrimaryAttackParams attackParams )
{
	entity ownerPlayer = weapon.GetWeaponOwner()

	Assert( IsValid( ownerPlayer) && ownerPlayer.IsPlayer() )

	if ( IsValid( ownerPlayer ) && ownerPlayer.IsPlayer() )
	{
		if ( ownerPlayer.GetCinematicEventFlags() & CE_FLAG_CLASSIC_MP_SPAWNING )
			return false

		if ( ownerPlayer.GetCinematicEventFlags() & CE_FLAG_INTRO )
			return false
	}

	PlayerUsedOffhand( ownerPlayer, weapon )

	#if SERVER

	if( weapon.HasMod( "cloak_field" ) )
	{
		entity deployable = ThrowDeployable( weapon, attackParams, DEPLOYABLE_THROW_POWER, OnDeployableCloakfieldPlanted )
		SendHudMessage(ownerPlayer, "Deployed Cloak Field\nSpecial Cloak in Cooldown", -1, -0.35, 255, 255, 100, 255, 0, 3, 0)
		thread CheckModdedCloak( weapon, "cloak_field", CLOAK_FIELD_COOLDOWN_TIME )
	}
	else if( weapon.HasMod( "cloak_drone" ) )
	{
		entity deployable = ThrowDeployable( weapon, attackParams, DEPLOYABLE_THROW_POWER, OnCloakDroneReleased )
		SendHudMessage(ownerPlayer, "Deployed Cloak Drone\nSpecial Cloak in Cooldown", -1, -0.35, 255, 255, 100, 255, 0, 3, 0)
		thread CheckModdedCloak( weapon, "cloak_drone", CLOAK_DRONE_COOLDOWN_TIME )
	}
	else
	{
		float duration = weapon.GetWeaponSettingFloat( eWeaponVar.fire_duration )
		EnableCloak( ownerPlayer, duration )
		#if BATTLECHATTER_ENABLED
			TryPlayWeaponBattleChatterLine( ownerPlayer, weapon )
		#endif
	}
		//ownerPlayer.Signal( "PlayerUsedAbility" )
	#endif

	return weapon.GetWeaponSettingInt( eWeaponVar.ammo_min_to_fire )
}

void function OnDeployableCloakfieldPlanted( entity projectile )
{
	#if SERVER
	DeployCloakfield( projectile )
	#endif
}

void function OnCloakDroneReleased( entity projectile )
{
	#if SERVER
	entity drone = SpawnCloakDrone( projectile.GetTeam(), projectile.GetOrigin(), < 0,0,0 >, < 0,0,0 >, projectile.GetOwner() )
	thread AfterTimeDestroyDrone( drone, CLOAK_DRONE_LIFETIME )
	projectile.GrenadeExplode( < 0,0,20 > )
	#endif
}

#if SERVER
void function CheckModdedCloak( entity weapon, string mod, float cooldown )
{
	int offhandslot = -1
	if( !weapon.HasMod( mod ) )
		return
	entity weaponOwner = weapon.GetWeaponOwner()
	weaponOwner.EndSignal( "OnDeath" )
	weaponOwner.EndSignal( "OnDestroy" )
	foreach( entity offhand in weaponOwner.GetOffhandWeapons() )
	{
		offhandslot += 1
		if( offhandslot == OFFHAND_EQUIPMENT )
			offhandslot += 1
		if( offhand.GetWeaponClassName() == "mp_ability_cloak" )
		{

			weaponOwner.TakeWeaponNow( offhand.GetWeaponClassName() )
			break
		}
	}

	wait cooldown

	if( IsAlive(weaponOwner) && IsValid(weaponOwner) )
	{
		foreach( entity offhand in weaponOwner.GetOffhandWeapons() )
		{
			if( offhand.GetWeaponClassName() == "mp_ability_cloak" )
			{
				print( "Failed to refilled Cloakfield" )
				return
			}
		}
		if( !IsValid( weaponOwner.GetOffhandWeapon(offhandslot) ) )
		{
			SendHudMessage(weaponOwner, "Special Cloak Recharged", -1, -0.35, 255, 255, 100, 255, 0, 3, 0)
			weaponOwner.GiveOffhandWeapon( "mp_ability_cloak", offhandslot, [mod] )
		}
		else
			print( "Failed to refilled Cloakfield" )
	}
}

void function AfterTimeDestroyDrone( entity drone, float delay )
{
	wait delay
	if( IsValid( drone ) )
		drone.SetHealth( 0 )
}
#endif