global function OnWeaponPrimaryAttack_ability_heal

global const STIM_EFFECT_SEVERITY_OCTANE = 0.25
const OCTANE_STIM_DAMAGE = 20

const JUMP_PAD_LIFETIME = 15

var function OnWeaponPrimaryAttack_ability_heal( entity weapon, WeaponPrimaryAttackParams attackParams )
{
	entity ownerPlayer = weapon.GetWeaponOwner()
	Assert( IsValid( ownerPlayer) && ownerPlayer.IsPlayer() )
	if ( IsValid( ownerPlayer ) && ownerPlayer.IsPlayer() )
	{
		if ( ownerPlayer.GetCinematicEventFlags() & CE_FLAG_CLASSIC_MP_SPAWNING )
			return false

		if ( ownerPlayer.GetCinematicEventFlags() & CE_FLAG_INTRO )
			return false
	}

	float duration = weapon.GetWeaponSettingFloat( eWeaponVar.fire_duration )
	if( weapon.HasMod("jump_pad") )
	{
		entity deployable = ThrowDeployable( weapon, attackParams, DEPLOYABLE_THROW_POWER, OnJumpPadPlanted )
		#if SERVER
		SendHudMessage(ownerPlayer, "Deployed Jump Pad", -1, -0.35, 255, 255, 100, 255, 0, 3, 0)
		#endif
		if ( deployable )
		{
			entity player = weapon.GetWeaponOwner()

			#if SERVER
			string projectileSound = GetGrenadeProjectileSound( weapon )
			if ( projectileSound != "" )
				EmitSoundOnEntity( deployable, projectileSound )

			weapon.w.lastProjectileFired = deployable
			#endif
		}
	}
	else if( weapon.HasMod("octane_stim") )
	{
		StimPlayer( ownerPlayer, duration, STIM_EFFECT_SEVERITY_OCTANE )
		#if SERVER
		thread OctaneStimThink( ownerPlayer, duration )
		#endif
	}
	else if( weapon.HasMod( "bc_super_stim" ) && weapon.HasMod( "dev_mod_low_recharge" ) )
		StimPlayer( ownerPlayer, duration, 0.15 )
	else
		StimPlayer( ownerPlayer, duration )

	PlayerUsedOffhand( ownerPlayer, weapon )

#if SERVER
#if BATTLECHATTER_ENABLED
	TryPlayWeaponBattleChatterLine( ownerPlayer, weapon )
#endif //
#else //
	Rumble_Play( "rumble_stim_activate", {} )
#endif //

	return weapon.GetWeaponSettingInt( eWeaponVar.ammo_min_to_fire )
}

void function OnJumpPadPlanted( entity projectile )
{
	#if SERVER
	Assert( IsValid( projectile ) )
	vector origin = projectile.GetOrigin()

	vector endOrigin = origin - Vector( 0.0, 0.0, 32.0 )
	vector surfaceAngles = projectile.proj.savedAngles
	vector oldUpDir = AnglesToUp( surfaceAngles )

	TraceResults traceResult = TraceLine( origin, endOrigin, [], TRACE_MASK_SOLID, TRACE_COLLISION_GROUP_NONE )
	if ( traceResult.fraction < 1.0 )
	{
		vector forward = AnglesToForward( projectile.proj.savedAngles )
		surfaceAngles = AnglesOnSurface( traceResult.surfaceNormal, forward )

		vector newUpDir = AnglesToUp( surfaceAngles )
		if ( DotProduct( newUpDir, oldUpDir ) < 0.55 )
			surfaceAngles = projectile.proj.savedAngles
	}

	projectile.SetAngles( surfaceAngles )

	DeployJumpPad( projectile, origin, surfaceAngles )
	#endif
}

#if SERVER
void function DeployJumpPad( entity projectile, vector origin, vector angles )
{
	#if SERVER
	int team = projectile.GetTeam()
	entity tower = CreateEntity( "prop_dynamic" )
	tower.SetModel( $"models/weapons/sentry_shield/sentry_shield_proj.mdl" )
	tower.SetOrigin( origin )
	tower.SetAngles( angles )
	tower.kv.modelscale = 4

	thread JumpPadThink( tower, JUMP_PAD_LIFETIME )

	thread CleanupJumpPad( tower, projectile, JUMP_PAD_LIFETIME )
	#endif
}

void function JumpPadThink( entity tower, float duration )
{
	entity trigger = CreateTriggerRadiusMultiple( tower.GetOrigin(), 64, [], TRIG_FLAG_PLAYERONLY | TRIG_FLAG_NO_PHASE_SHIFT, 24, 0 )
	SetTeam( trigger, tower.GetTeam() )

	AddCallback_ScriptTriggerEnter( trigger, OnJumpPadTriggerEnter )

	ScriptTriggerSetEnabled( trigger, true )

	OnThreadEnd(
		function(): ( trigger ){
			trigger.Destroy()
		}
	)

	wait duration
}

void function OnJumpPadTriggerEnter( entity trigger, entity ent )
{
	thread GiveJumpPadEffect( ent )
}

void function GiveJumpPadEffect( entity player )
{
	player.SetVelocity( < player.GetVelocity().x *1.5, player.GetVelocity().y *1.5, 700 >  )
	for( int i = 0; i < 5; i++ )
		EmitSoundOnEntity( player, "Boost_Card_SentryTurret_Deployed_3P" )

	int attachmentIndex = player.LookupAttachment( "CHESTFOCUS" )

	entity fx = StartParticleEffectOnEntity_ReturnEntity( player, GetParticleSystemIndex( $"P_titan_sniper1" ), FX_PATTACH_POINT_FOLLOW, attachmentIndex )
	fx.SetOwner( player )
	fx.kv.VisibilityFlags = (ENTITY_VISIBLE_TO_FRIENDLY | ENTITY_VISIBLE_TO_ENEMY)

	OnThreadEnd(
		function(): ( fx ){
			if( IsValid(fx) )
				EffectStop(fx)
		}
	)

	wait 2
}

void function CleanupJumpPad( entity tower, entity projectile, float delay )
{
	wait delay
	if( IsValid(tower) )
		tower.Destroy()
	if( IsValid(projectile) )
		projectile.GrenadeExplode(< 0,0,10 >)
}

void function OctaneStimThink( entity player, float duration )
{
	player.EndSignal( "OnDeath" )
	player.EndSignal( "OnDestroy" )
	float startTime = Time()
	player.TakeDamage( player.GetHealth() - OCTANE_STIM_DAMAGE < 1 ? 1 : OCTANE_STIM_DAMAGE, player, player, { damageSourceId = eDamageSourceId.bleedout } )
	while( true )
	{
		wait 0.1
		if( IsValid( player ) )
			player.p.lastDamageTime = Time()
		if( Time() >= startTime + duration )
		{
			if( IsValid( player ) )
				player.p.lastDamageTime = Time() - 5
			return
		}
	}
}
#endif
